/*
        This is a sample plugin. It illustrates colorizing

          by coloring the current loop in yellow

*/

#include <set>
#include <ida.hpp>
#include <idp.hpp>
#include <gdl.hpp>
#include <name.hpp>
#include <frame.hpp>
#include <loader.hpp>
#include <struct.hpp>
#include <kernwin.hpp>
#include <allins.hpp>

// colors are unsaturated to make the listing more readable
static const bgcolor_t YELLOW = 0xC0FFFF;

// the current screen address
static ea_t displayed_ea = BADADDR;
static bool must_refresh = true;

// set of current loop basic blocks
static func_t *current_pfn = NULL;
static intset_t current_loop;

// set of colored instructions
typedef std::set<ea_t> easet_t;
static easet_t colored;

// we will need to refresh idaviews
void (idaapi*idaview_marker)(ea_t ea);

//--------------------------------------------------------------------------
// define these functions to shut the linker. we do not use these functions
// anyway
void idaapi flow_chart_t::print_graph_attributes(FILE * /*fp*/) const {}
bool idaapi flow_chart_t::print_node(FILE * /*fp*/, int /*n*/) const { return false; }
bool idaapi flow_chart_t::print_edge(FILE * /*fp*/, int /*i*/, int /*j*/) const { return false; }

//--------------------------------------------------------------------------
// the flow chart has no information about predecessors
// we have to calculate it outselves
typedef qvector<intvec_t> preds_t;

static void calc_predecessors(const flow_chart_t &fc, preds_t &preds)
{
  int nblocks = fc.blocks.size();
  preds.resize(nblocks);
  for ( int i=0; i < nblocks; i++ )
  {
    const intvec_t &succ = fc.blocks[i].succ;
    for ( int j=0; j < succ.size(); j++ )
      preds[succ[j]].push_back(i);
  }
}

//--------------------------------------------------------------------------
// walk forward the graph starting at the given node
static void find_successors(const flow_chart_t &fc, int node, intset_t &set)
{
  if ( set.insert(node).second ) // inserted successfully => was not present
  {
    int ns = fc.nsucc(node); // number of sucessors
    for ( int i=0; i < ns; i++ )
      find_successors(fc, fc.succ(node, i), set);
  }
}

//--------------------------------------------------------------------------
// walk backward the graph starting at the given node
static void find_predecessors(const preds_t &preds, int node, intset_t &set)
{
  if ( set.insert(node).second ) // inserted successfully => was not present
  {
    intvec_t node_preds = preds[node];
    int np = node_preds.size(); // number of predecessors
    for ( int i=0; i < np; i++ )
      find_predecessors(preds, node_preds[i], set);
  }
}

//--------------------------------------------------------------------------
// find out the basic block numbers of the current loop
static intset_t gather_loop_blocks(const flow_chart_t &fc, ea_t ea)
{
  intset_t set; // return value

  // find the current basic block
  int i;
  for ( i=0; i < fc.blocks.size(); i++ )
  {
    if ( fc.blocks[i].contains(ea) )
      break;
  }
  // found the basic block?
  // if not, may be the cursor is on a data item?
  if ( i < fc.blocks.size() )
  {
    // we have a loop if there is a path in the graph
    // starting at the current node and ending at it
    intset_t successors;
    find_successors(fc, i, successors);

    // the flow chart has no information about node predecessors
    // we have to calculate them ourselves
    preds_t preds;
    calc_predecessors(fc, preds);
    intset_t predecessors;
    find_predecessors(preds, i, predecessors);
    // intersection of these two sets are the loop nodes
    for ( intset_t::iterator p=successors.begin();
          p != successors.end();
          ++p )
    {
      if ( predecessors.find(*p) != predecessors.end() )
        set.insert(*p);
    }
    // if there is only one node, there might no loop
    if ( set.size() == 1 )
    { // the current node refers to itself?
      if ( !fc.blocks[i].succ.has(i) )
        set.clear(); // no, this is not a loop
    }
  }
  return set;
}

//--------------------------------------------------------------------------
// color all instructions of the basic block with the given color
static void set_bb_color(const area_t &a, bgcolor_t color)
{
  // for all instructions
  for ( ea_t ea=a.startEA; ea < a.endEA; ea=nextthat(ea, a.endEA, f_isCode, NULL) )
  {
    set_item_color(ea, color);
    if ( color != DEFCOLOR )
      colored.insert(ea);
    mark_idaview_for_refresh(ea);
  }
}

//--------------------------------------------------------------------------
// color all basic blocks in 'loop'
static void set_item_colors(const flow_chart_t &fc, const intset_t &loop, bgcolor_t color)
{
  for ( intset_t::const_iterator p=loop.begin(); p != loop.end(); ++p )
    set_bb_color(fc.blocks[*p], color);
}

//--------------------------------------------------------------------------
// decolorise an old loop
static void decolorize(void)
{
  if ( !colored.empty() )
  {
    for ( easet_t::iterator p=colored.begin(); p != colored.end(); ++p )
    {
      set_item_color(*p, DEFCOLOR);
      mark_idaview_for_refresh(*p);
    }
    colored.clear();
  }
}

//--------------------------------------------------------------------------
// the current address has been changed. refresh the listing
static void refresh_colors(ea_t ea)
{
  must_refresh = false;
  func_t *pfn = get_func(ea);
  if ( pfn == NULL )
  {
    decolorize();
    current_pfn = NULL;
  }
  else
  {

    // build flow chart of the current function
    flow_chart_t fc("", pfn, BADADDR, BADADDR, false);

    intset_t new_loop = gather_loop_blocks(fc, ea);
    if ( pfn != current_pfn || new_loop != current_loop )
    {
      decolorize();
      set_item_colors(fc, new_loop, YELLOW);
      current_loop.swap(new_loop);
      current_pfn = pfn;
    }
  }
}

//--------------------------------------------------------------------------
// We hook to IDP event to detect the database modifications
static int idaapi idp_callback(void *, int notification_code, va_list va)
{
  switch ( notification_code)
  {
    case processor_t::undefine:
    case processor_t::make_code:
    case processor_t::make_data:
      {
        ea_t ea = va_arg(va, ea_t);
        if ( displayed_ea == ea )  // the currently displayed address is being
          must_refresh = true;     // changed. it makes no sense to refresh
                                   // our window immediately since the changes
                                   // are not in the database yet. we just raise
                                   // a flag here and will check it in the ui
                                   // callback
      }
      break;
  }
  return 0;
}

//--------------------------------------------------------------------------
// We hook to user-interface events to detect the cursor movements
static int idaapi ui_callback(void *, int notification_code, va_list)
{
  switch ( notification_code )
  {
    case ui_setstate:
      // the 'setstate' event occurs very often - periodically when
      // ida is idle. we will use it to detect the cursor movements
      {
        ea_t ea = get_screen_ea(); // current cursor address
        if ( ea != displayed_ea || must_refresh )
        {
          displayed_ea = ea;
          refresh_colors(ea);
        }
      }
      break;
    case ui_refresh:
      // when all disassembly listings are refreshed, we refresh our window too
      refresh_colors(get_screen_ea());
      break;
  }

  return 0;
}

//--------------------------------------------------------------------------
int idaapi init(void)
{
  // initialize idaview_marker() pointer
  setup_idaview_marker();
  // hook events about database modifications
  hook_to_notification_point(HT_IDP, idp_callback, NULL);
  // hook events about user interface modifications
  hook_to_notification_point(HT_UI, ui_callback, NULL);
  return PLUGIN_KEEP;
}

//--------------------------------------------------------------------------
void idaapi term(void)
{
  // remove our colors
  decolorize();
  unhook_from_notification_point(HT_IDP, idp_callback, NULL);
  unhook_from_notification_point(HT_UI, ui_callback, NULL);
}

//--------------------------------------------------------------------------
void idaapi run(int)
{
  // since the plugin is fully automatic, there is nothing to do
  warning("The colorizer plugin is fully automatic");
}

//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  PLUGIN_HIDE,          // the plugin won't be visible in the menu
  init,                 // initialize
  term,                 // terminate. this pointer may be NULL.
  run,                  // invoke plugin
  NULL,                 // long comment about the plugin
  NULL,                 // multiline help about the plugin
  "Colorize loops",     // the preferred short name of the plugin
  NULL                  // the preferred hotkey to run the plugin
};
